/***************************************************************************
 *   Copyright (C) 2007  Antony Martin <martin.antony@yahoo.fr>            *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or         *
 *   modify it under the terms of the GNU General Public License           *
 *   as published by the Free Software Foundation; either version 2        *
 *   of the License, or (at your option) any later version.                *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA          *
 ***************************************************************************/

/* Cree le : 6 mars 2006
   derniere modification le 19/10/2007 */

/**
A implementer :
_______________
\p ---> gestion des parametres de shader a envoyer a chaque fois
        (variable... variable :D), prevoir d'utiliser un pointeur sur la
        variable, et envoyer son contenu a chaque utilisation du shader
        (ex. un vecteur de lumiere)

*/

#include <SCE/SCEMinimal.h>

#include <SCE/utils/SCEString.h>
#include <SCE/utils/SCEMedia.h>
#include <SCE/utils/SCEResources.h>

#include <SCE/interface/SCEShaders.h>


typedef void (*SCE_FShaderDelete)(SCE_SShader*);
static void SCE_Shader_DeleteASM (SCE_SShader*);
#ifdef _SCE_USE_CG
static void SCE_Shader_DeleteCG (SCE_SShader*);
#endif
static void SCE_Shader_DeleteGLSL (SCE_SShader*);

typedef int (*SCE_FShaderBuild)(SCE_SShader*);
static int SCE_Shader_BuildASM (SCE_SShader*);
#ifdef _SCE_USE_CG
static int SCE_Shader_BuildCG (SCE_SShader*);
#endif
static int SCE_Shader_BuildGLSL (SCE_SShader*);

/*******/
typedef int (*SCE_FShaderGetIndex)(SCE_SShader*, int, const char*);
static int SCE_Shader_GetIndexASM (SCE_SShader*, int, const char*);
#ifdef _SCE_USE_CG
static int SCE_Shader_GetIndexCG (SCE_SShader*, int, const char*);
#endif
static int SCE_Shader_GetIndexGLSL (SCE_SShader*, int, const char*);

/*******/
typedef void (*SCE_FShaderParam)(int, const char*, int);
static void SCE_Shader_ParamASM (int, const char*, int);
#ifdef _SCE_USE_CG
static void SCE_Shader_ParamCG (int, const char*, int);
#endif
static void SCE_Shader_ParamGLSL (int, const char*, int);

typedef void (*SCE_FShaderParamf)(int, const char*, float);
static void SCE_Shader_ParamfASM (int, const char*, float);
#ifdef _SCE_USE_CG
static void SCE_Shader_ParamfCG (int, const char*, float);
#endif
static void SCE_Shader_ParamfGLSL (int, const char*, float);

typedef void (*SCE_FShaderParamfv)(int, const char*, size_t, float*);
static void SCE_Shader_ParamfvASM (int, const char*, size_t, float*);
#ifdef _SCE_USE_CG
static void SCE_Shader_ParamfvCG (int, const char*, size_t, float*);
#endif
static void SCE_Shader_ParamfvGLSL (int, const char*, size_t, float*);

/*******/
typedef void (*SCE_FShaderSetParam)(int, int);
static void SCE_Shader_SetParamASM (int, int);
#ifdef _SCE_USE_CG
static void SCE_Shader_SetParamCG (int, int);
#endif
static void SCE_Shader_SetParamGLSL (int, int);

typedef void (*SCE_FShaderSetParamf)(int, float);
static void SCE_Shader_SetParamfASM (int, float);
#ifdef _SCE_USE_CG
static void SCE_Shader_SetParamfCG (int, float);
#endif
static void SCE_Shader_SetParamfGLSL (int, float);

typedef void (*SCE_FShaderSetParamfv)(int, size_t, float*);
static void SCE_Shader_SetParamfvASM (int, size_t, float*);
#ifdef _SCE_USE_CG
static void SCE_Shader_SetParamfvCG (int, size_t, float*);
#endif
static void SCE_Shader_SetParamfvGLSL (int, size_t, float*);
/*******/
typedef void (*SCE_FShaderUse)(SCE_SShader*);
static void SCE_Shader_UseASM (SCE_SShader*);
#ifdef _SCE_USE_CG
static void SCE_Shader_UseCG (SCE_SShader*);
#endif
static void SCE_Shader_UseGLSL (SCE_SShader*);


static const SCE_FShaderDelete Delete[3] =
{
    SCE_Shader_DeleteASM,
#ifdef _SCE_USE_CG
    SCE_Shader_DeleteCG,
#else
    NULL,
#endif
    SCE_Shader_DeleteGLSL
};

/*******/
static const SCE_FShaderBuild Build[3] =
{
    SCE_Shader_BuildASM,
#ifdef _SCE_USE_CG
    SCE_Shader_BuildCG,
#else
    NULL,
#endif
    SCE_Shader_BuildGLSL
};

/*******/
static const SCE_FShaderGetIndex GetIndex[3] =
{
    SCE_Shader_GetIndexASM,
#ifdef _SCE_USE_CG
    SCE_Shader_GetIndexCG,
#else
    NULL,
#endif
    SCE_Shader_GetIndexGLSL
};

/*******/
static const SCE_FShaderParam Param[3] =
{
    SCE_Shader_ParamASM,
#ifdef _SCE_USE_CG
    SCE_Shader_ParamCG,
#else
    NULL,
#endif
    SCE_Shader_ParamGLSL
};

static const SCE_FShaderParamf Paramf[3] =
{
    SCE_Shader_ParamfASM,
#ifdef _SCE_USE_CG
    SCE_Shader_ParamfCG,
#else
    NULL,
#endif
    SCE_Shader_ParamfGLSL
};

static const SCE_FShaderParamfv Paramfv[3] =
{
    SCE_Shader_ParamfvASM,
#ifdef _SCE_USE_CG
    SCE_Shader_ParamfvCG,
#else
    NULL,
#endif
    SCE_Shader_ParamfvGLSL
};

/*******/
static const SCE_FShaderSetParam SetParam[3] =
{
    SCE_Shader_SetParamASM,
#ifdef _SCE_USE_CG
    SCE_Shader_SetParamCG,
#else
    NULL,
#endif
    SCE_Shader_SetParamGLSL
};

static const SCE_FShaderSetParamf SetParamf[3] =
{
    SCE_Shader_SetParamfASM,
#ifdef _SCE_USE_CG
    SCE_Shader_SetParamfCG,
#else
    NULL,
#endif
    SCE_Shader_SetParamfGLSL
};

static const SCE_FShaderSetParamfv SetParamfv[3] =
{
    SCE_Shader_SetParamfvASM,
#ifdef _SCE_USE_CG
    SCE_Shader_SetParamfvCG,
#else
    NULL,
#endif
    SCE_Shader_SetParamfvGLSL
};

/*******/
static const SCE_FShaderUse Use[3] =
{
    SCE_Shader_UseASM,
#ifdef _SCE_USE_CG
    SCE_Shader_UseCG,
#else
    NULL,
#endif
    SCE_Shader_UseGLSL
};


/* identifiant de type du media manager */
static int shader_mediatypeid = 0;

static int sce_shd_enabled = SCE_FALSE;

static SCE_SShader *used = NULL;


/* revise le 19/10/2007 */
int SCE_Init_Shader (void)
{
    static int is_init = SCE_FALSE;

    SCE_btstart ();
    if (!is_init)
    {
        shader_mediatypeid = SCE_Media_GenTypeID();
        SCE_Media_RegisterLoader (shader_mediatypeid, 0,
                                  ".glsl .vert .frag .asm .shd .vs .ps .fs .vsh"
                                  " .psh .fsh"
#ifdef _SCE_USE_CG
                                  " .cg .vcg .pcg .fcg .cgvs .cgps"
#endif
                                  ,SCE_Shader_LoadSourceFromFile);

        is_init = SCE_TRUE;
        sce_shd_enabled = SCE_TRUE;
    }

    SCE_btend ();
    return SCE_OK;
}

/* revise le 18/10/2007 */
void SCE_Quit_Shader (void)
{
    sce_shd_enabled = SCE_TRUE;
    used = NULL;
}

int SCE_Shader_MediaTypeID (void)
{
    return shader_mediatypeid;
}


SCE_SShader* SCE_Shader_Create (int type)
{
    SCE_SShader *shader = NULL;

    SCE_btstart ();
    shader = SCE_malloc (sizeof *shader);
    if (!shader)
    {
        Logger_LogSrc ();
        SCE_btend ();
        return NULL;
    }

    shader->p_glsl = NULL;
    shader->v = shader->p = NULL;

    shader->type = type;
    shader->ready = SCE_FALSE;
    shader->srcs = NULL;
    shader->vs_source = shader->ps_source = NULL;
    shader->vs_addsrc = shader->ps_addsrc = NULL;

    SCE_btend ();
    return shader;
}


static void SCE_Shader_DeleteASM (SCE_SShader *shader)
{
    SCE_btstart ();
    SCE_CDeleteShaderASM (shader->v);
    SCE_CDeleteShaderASM (shader->p);
    SCE_btend ();
}
#ifdef _SCE_USE_CG
static void SCE_Shader_DeleteCG (SCE_SShader *shader)
{
    SCE_btstart ();
    SCE_CDeleteShaderCG (shader->v);
    SCE_CDeleteShaderCG (shader->p);
    SCE_btend ();
}
#endif
static void SCE_Shader_DeleteGLSL (SCE_SShader *shader)
{
    SCE_btstart ();
    SCE_CDeleteProgram (shader->p_glsl);
    SCE_CDeleteShaderGLSL (shader->v);
    SCE_CDeleteShaderGLSL (shader->p);
    SCE_btend ();
}
/* revise le 19/10/2007 */
void SCE_Shader_Delete (SCE_SShader *shader)
{
    SCE_btstart ();
    if (shader)
    {
        Delete[shader->type] (shader);
        
        SCE_free (shader->vs_addsrc);
        SCE_free (shader->ps_addsrc);

        if (SCE_Resource_Free (shader->srcs))
        {
            SCE_free (shader->srcs[0]);
            SCE_free (shader->srcs[1]);
            SCE_free (shader->srcs);
        }

        SCE_free (shader);
    }
    SCE_btend ();
}

int SCE_Shader_GetLanguage (SCE_SShader *shader)
{
    return shader->type;
}

int SCE_Shader_GetType (SCE_SShader *shader)
{
    if ((shader->ps_source != NULL || shader->ps_addsrc != NULL) &&
        (shader->vs_source != NULL || shader->vs_addsrc != NULL) )
        return SCE_UNKNOWN_SHADER;
    else if (shader->ps_source != NULL || shader->ps_addsrc != NULL)
        return SCE_PIXEL_SHADER;
    else
        return SCE_VERTEX_SHADER;
}


/* retourne le type d'un shader d'apres son extension */
static void SCE_Shader_SearchTypes (const char *ext, int *type)
{
    SCE_btstart ();
    #define SCE_SHADER_FOR(str, val, val2)\
    if (SCE_String_Cmp (str, ext, 0))\
    {\
        type[0] = val;\
        type[1] = val2;\
    }\
    else

    /* NOTE: c'est bien statique tout ça... il faudrait rajouter une
             fonctionnalite de stockage d'extensions */
    SCE_SHADER_FOR (".glsl", SCE_GLSL_SHADER, SCE_UNKNOWN_SHADER)
    SCE_SHADER_FOR (".vert", SCE_GLSL_SHADER, SCE_VERTEX_SHADER)
    SCE_SHADER_FOR (".frag", SCE_GLSL_SHADER, SCE_PIXEL_SHADER)

    SCE_SHADER_FOR (".asm", SCE_ASM_SHADER, SCE_UNKNOWN_SHADER)
    SCE_SHADER_FOR (".shd", SCE_ASM_SHADER, SCE_UNKNOWN_SHADER)
    SCE_SHADER_FOR (".vs", SCE_ASM_SHADER, SCE_VERTEX_SHADER)
    SCE_SHADER_FOR (".ps", SCE_ASM_SHADER, SCE_PIXEL_SHADER)
    SCE_SHADER_FOR (".fs", SCE_ASM_SHADER, SCE_PIXEL_SHADER)
    SCE_SHADER_FOR (".vsh", SCE_ASM_SHADER, SCE_VERTEX_SHADER)
    SCE_SHADER_FOR (".psh", SCE_ASM_SHADER, SCE_PIXEL_SHADER)
    SCE_SHADER_FOR (".fsh", SCE_ASM_SHADER, SCE_PIXEL_SHADER)

    #ifdef _SCE_USE_CG
    SCE_SHADER_FOR (".cg", SCE_CG_SHADER, SCE_UNKNOWN_SHADER)
    SCE_SHADER_FOR (".vcg", SCE_CG_SHADER, SCE_VERTEX_SHADER)
    SCE_SHADER_FOR (".pcg", SCE_CG_SHADER, SCE_PIXEL_SHADER)
    SCE_SHADER_FOR (".fcg", SCE_CG_SHADER, SCE_PIXEL_SHADER)
    SCE_SHADER_FOR (".cgvs", SCE_CG_SHADER, SCE_VERTEX_SHADER)
    SCE_SHADER_FOR (".cgps", SCE_CG_SHADER, SCE_PIXEL_SHADER)
    #endif
    {
        Logger_Log (SCE_INVALID_ARG);
        Logger_LogMsg ("'%s' is not a valid shader source extension", ext);
        type[0] = type[1] = SCE_UNKNOWN_SHADER;
    }

    #undef SCE_SHADER_FOR
    SCE_btend ();
}


/* NOTE: a deplacer */
/* positionne le curseur de fp sur la prochaine occurrence a str
   retourne 1 si la fonction a trouve une occurrence */
static int SCE_Shader_SetPosFile (FILE *fp, const char *str, int at_end)
{
    size_t i = 0;
    long cur = 0, curo = 0;
    char buf[256] = {0};
    size_t lenstr = 0;

    SCE_btstart ();
    lenstr = strlen (str);
    curo = ftell (fp);

    /* NOTE: algorithme a ameliorer...
             creer une fonction qui propose de relever toutes les positions
             de toutes les chaines passees en parametres (params variables) */
    while (1)
    {
        fseek (fp, curo+cur, SEEK_SET);
        cur++;

        for (i=0; i<lenstr; i++)
        {
            buf[i] = fgetc (fp);
            
            if (buf[i] == EOF) /* fin du fichier */
            {
                fseek (fp, curo, SEEK_SET);
                SCE_btend ();
                return 0;
            }
        }

        if (SCE_String_Cmp (str, buf, 0))
        {
            /* on se replace au debut */
            if (!at_end)
                fseek (fp, -lenstr, SEEK_CUR);
            SCE_btend ();
            return 1;
        }
    }

    fseek (fp, curo, SEEK_SET);

    SCE_btend ();
    return 0;
}

/* charge du texte a partir d'un fichier, jusqu'a end */
static char* SCE_Shader_LoadSource (FILE *fp, long end)
{
    char *src = NULL;
    long len = ftell (fp);
    long curpos;

    SCE_btstart ();
    /* calcul de la taille total de la chaine */
    if (end == EOF)
    {
        /* on calcul la distance qu'il y a jusqu'a la fin du fichier */
        curpos = len;
        fseek (fp, 0, SEEK_END);
        len = ftell (fp) - len;
        fseek (fp, curpos, SEEK_SET);
    }
    else
        len = end - len;

    /* dans le cas ou la zone serait vide, aucun code de shader du type prefixe
       n'est present a cet endroit, mais pas la peine d'en faire un fromage */
    if (len <= 0)
    {
        SCE_btend ();
        return NULL;
    }

    /* allocation */
    src = SCE_malloc (len+1);
    if (!src)
    {
        Logger_LogSrc ();
        SCE_btend ();
        return NULL;
    }
    src[len] = '\0';

    /* lecture */
    fread (src, 1, len, fp);

    SCE_btend ();
    return src;
}

void* SCE_Shader_LoadSourceFromFile (FILE *fp, const char *fname, void *uusd)
{
    char **srcs = NULL;
    long vpos, ppos, vend, pend;
    int type[2];

    SCE_btstart ();
    /* inutilise */
    uusd = NULL;

    srcs = SCE_malloc (2 * sizeof *srcs);
    if (!srcs)
    {
        Logger_LogSrc ();
        SCE_btend ();
        return NULL;
    }
    srcs[0] = srcs[1] = NULL;

    /* recherche du type du shader */
    SCE_Shader_SearchTypes (SCE_String_GetExt ((char*)fname), type);
    if (type[1] == SCE_UNKNOWN_SHADER)
    {
        /* le fichier contient le code du vertex et du pixel shader */
        /* recuperation de la position du code du vertex shader */
        SCE_Shader_SetPosFile (fp, "[vertex shader]", SCE_FALSE);
        vpos = ftell (fp);
        vend = vpos + 15;   /* 16 = strlen("[vertex shader]"); */
        rewind (fp);

        /* recuperation de la position du code du pixel shader */
        SCE_Shader_SetPosFile (fp, "[pixel shader]", SCE_FALSE);
        ppos = ftell (fp);
        pend = ppos + 14;   /* 15 = strlen("[pixel shader]"); */

        /* lecture du code du vertex shader */
        fseek (fp, vend, SEEK_SET);
        srcs[0] = SCE_Shader_LoadSource (fp, (vpos > ppos) ? EOF : ppos);

        /* lecture du code du pixel shader */
        fseek (fp, pend, SEEK_SET);
        srcs[1] = SCE_Shader_LoadSource (fp, (ppos > vpos) ? EOF : vpos);
    }
    else
    {
        /* le fichier contient le code de type[1] */
        srcs[0] = SCE_Shader_LoadSource (fp, EOF);

        if (type[1] == SCE_PIXEL_SHADER)
        {
            srcs[1] = srcs[0];
            srcs[0] = NULL;
        }
    }

    SCE_btend ();
    return srcs;
}

SCE_SShader* SCE_Shader_CreateFromFile (const char *vname, const char *pname)
{
    SCE_SShader *shader = NULL;
    char **srcs = NULL;
    char *vsource = NULL, *psource = NULL;
    int type[2];
    int ttemp = SCE_UNKNOWN_SHADER;

    SCE_btstart ();
    if (vname)
    {
        SCE_Shader_SearchTypes (SCE_String_GetExt ((char*)vname), type);

        srcs = SCE_Resource_Load (vname, NULL, NULL);
        if (!srcs)
        {
            Logger_LogSrc ();
            SCE_btend ();
            return NULL;
        }

        if (srcs[0] && srcs[1])
        {
            /* on vient de charger un fichier contenant les deux codes */
            pname = NULL;
        }
        
        vsource = srcs[0];
        psource = srcs[1];
        ttemp = type[0];
    }

    if (pname)
    {
        SCE_Shader_SearchTypes (SCE_String_GetExt ((char*)pname), type);

        srcs = SCE_Resource_Load (pname, NULL, NULL);
        if (!srcs)
        {
            Logger_LogSrc ();
            SCE_btend ();
            return NULL;
        }

        if(srcs[0] && srcs[1])
        {
            /* on vient de charger un fichier contenant les deux codes */
            vsource = srcs[0];
        }

        psource = srcs[1];

        if (ttemp != type[0] && vname)
        {
            /* le type du pixel shader differe de celui du vertex shader */

            if (SCE_Resource_Free (srcs))
            {
                SCE_free (vsource);
                SCE_free (psource);
                SCE_free (srcs);
            }

            Logger_Log (SCE_INVALID_ARG);
            Logger_LogMsg ("you can't load a %s vertex shader with"
                           " a %s pixel shader",
                           (ttemp == SCE_GLSL_SHADER) ? "GLSL" :
                           (ttemp == SCE_ASM_SHADER)  ? "ASM" : "Cg",
                           (type[0] == SCE_GLSL_SHADER) ? "GLSL" :
                           (type[0] == SCE_ASM_SHADER)  ? "ASM" : "Cg");
            SCE_btend ();
            return NULL;
        }
    }

    /* creation du shader */
    shader = SCE_Shader_Create (type[0]);
    if (!shader)
    {
        if (SCE_Resource_Free (srcs))
        {
            SCE_free (vsource);
            SCE_free (psource);
            SCE_free (srcs);
        }
        Logger_LogSrc ();
        SCE_btend ();
        return NULL;
    }

    shader->srcs = srcs;
    shader->vs_source = vsource;
    shader->ps_source = psource;

    SCE_btend ();
    return shader;
}


#define SCE_SHADER_BUILDFUNC(bign, add)\
static int SCE_Shader_Build##bign (SCE_SShader *shader)\
{\
    SCE_btstart ();\
    if (shader->vs_source)\
    {\
        shader->v = SCE_CCreateShader##bign (SCE_VERTEX_SHADER);\
        if (!shader->v)\
        {\
            Logger_LogSrc ();\
            SCE_btend ();\
            return SCE_ERROR;\
        }\
        SCE_CSetShader##bign##Source (shader->v, shader->vs_source);\
        if (SCE_CBuildShader##bign (shader->v) < 0)\
        {\
            SCE_CDeleteShader##bign (shader->v);\
            Logger_LogSrc ();\
            SCE_btend ();\
            return SCE_ERROR;\
        }\
    }\
\
    if (shader->ps_source)\
    {\
        shader->p = SCE_CCreateShader##bign (SCE_PIXEL_SHADER);\
        if (!shader->p)\
        {\
            Logger_LogSrc ();\
            SCE_btend ();\
            return SCE_ERROR;\
        }\
        SCE_CSetShader##bign##Source (shader->p, shader->ps_source);\
        if (SCE_CBuildShader##bign (shader->p) < 0)\
        {\
            SCE_CDeleteShader##bign (shader->p);\
            Logger_LogSrc ();\
            SCE_btend ();\
            return SCE_ERROR;\
        }\
    }\
    add\
    SCE_btend ();\
    return SCE_OK;\
}

SCE_SHADER_BUILDFUNC (ASM, )
#ifdef _SCE_USE_CG
SCE_SHADER_BUILDFUNC (CG, )
#endif
SCE_SHADER_BUILDFUNC
(GLSL, 
 shader->p_glsl = SCE_CCreateProgram ();
 if (!shader->p_glsl)
 {
     SCE_CDeleteShaderGLSL (shader->v);
     SCE_CDeleteShaderGLSL (shader->p);
     Logger_LogSrc ();
     SCE_btend ();
     return SCE_ERROR;
 }
 if (shader->v)
     SCE_CSetProgramShader (shader->p_glsl, shader->v, 1);
 if (shader->p)
     SCE_CSetProgramShader (shader->p_glsl, shader->p, 1);
 if (SCE_CBuildProgram (shader->p_glsl) < 0)
 {
     SCE_CDeleteProgram (shader->p_glsl);
     SCE_CDeleteShaderGLSL (shader->v);
     SCE_CDeleteShaderGLSL (shader->p);
     Logger_LogSrc ();
     SCE_btend ();
     return SCE_ERROR;
 }
)

/* revise le 19/10/2007 */
int SCE_Shader_Build (SCE_SShader *shader)
{
    SCE_btstart ();
    #ifdef __SCE_DEBUG__
    /* si le shader ne contient aucun code source */
    if (!shader->vs_source && !shader->ps_source &&
        !shader->vs_addsrc && !shader->ps_addsrc)
    {
        Logger_Log (SCE_INVALID_OPERATION);
        Logger_LogMsg ("this shader don't have a source code!");
        SCE_btend ();
        return SCE_ERROR;
    }
    #endif

    if (shader->vs_source || shader->vs_addsrc)
    {
        shader->vs_source = SCE_String_CatDup (shader->vs_addsrc, shader->vs_source);
        if (!shader->vs_source)
        {
            Logger_LogSrc ();
            SCE_btend ();
            return SCE_ERROR;
        }
    }
    if (shader->ps_source || shader->ps_addsrc)
    {
        shader->ps_source = SCE_String_CatDup (shader->ps_addsrc, shader->ps_source);
        if (!shader->ps_source)
        {
            Logger_LogSrc ();
            SCE_btend ();
            return SCE_ERROR;
        }
    }

    if (Build[shader->type] (shader) < 0)
    {
        Logger_LogSrc ();
        SCE_btend ();
        return SCE_ERROR;
    }

    shader->ready = SCE_TRUE;
    SCE_btend ();
    return SCE_OK;
}


/* revise le 19/10/2007 */
int SCE_Shader_AddSource (SCE_SShader *shader, int type, char *src)
{
    /* taille de la reallocation */
    size_t realen;
    int first_alloc = 1;
    char *addsrc = NULL;

    SCE_btstart ();
    if (!src)
    {
        SCE_btend ();
        return SCE_OK;
    }

    #ifdef __SCE_DEBUG__
    if (type != SCE_PIXEL_SHADER && type != SCE_VERTEX_SHADER &&
        type != SCE_UNKNOWN_SHADER)
    {
        Logger_Log (SCE_INVALID_ARG);
        SCE_btend ();
        return SCE_ERROR;
    }
    #endif

    addsrc =
    (type == SCE_PIXEL_SHADER) ? shader->ps_addsrc : shader->vs_addsrc;

    /* + 2 car, 1: retour chariot. 2: caractere de fin de chaine */
    realen = strlen(src) + 2;

    if (addsrc)
    {
        first_alloc = 0;
        realen += strlen (addsrc);
    }

    addsrc = SCE_realloc (addsrc, realen);
    if (!addsrc)
    {
        Logger_LogSrc ();
        SCE_btend ();
        return SCE_ERROR;
    }
    /* s'il s'agit de la premiere allocation, on initialise le contenu */
    if (first_alloc)
        memset (addsrc, '\0', realen);

    strcat (addsrc, src);

    /* affectation */
    if (type == SCE_PIXEL_SHADER)
        shader->ps_addsrc = addsrc;
    else
        shader->vs_addsrc = addsrc;

    SCE_btend ();
    return SCE_OK;
}


/**
 * \todo A coder...
 */
/* localise et renvoie un index au sein du code source d'un shader ASM */
static int SCE_Shader_GetASMIndex (const char *src, const char *name)
{
    src = name = NULL;
    return SCE_SHADER_BAD_INDEX;
}


static int SCE_Shader_GetIndexASM (SCE_SShader *s, int t, const char *n)
{
    /* on va effectuer une judicieuse recherche au sein du code source... */
    if (t == SCE_PIXEL_SHADER)
        return SCE_Shader_GetASMIndex (s->ps_source, n);
    else
        return SCE_Shader_GetASMIndex (s->vs_source, n);
}
#ifdef _SCE_USE_CG
static int SCE_Shader_GetIndexCG (SCE_SShader *s, int t, const char *n)
{
    if (t == SCE_PIXEL_SHADER)
        return (int)SCE_CGetShaderCGIndex (s->p, n);
    else
        return (int)SCE_CGetShaderCGIndex (s->v, n);
}
#endif
static int SCE_Shader_GetIndexGLSL (SCE_SShader *s, int t, const char *n)
{
    (void)t;
    return SCE_CGetProgramIndex (s->p_glsl, n);
}

/* revise le 19/10/2007 */
int SCE_Shader_GetIndex (SCE_SShader *shader, int type, const char *name)
{
    #ifdef __SCE_DEBUG__
    if (!shader || !name)
    {
        Logger_Log (SCE_INVALID_ARG);
        return SCE_SHADER_BAD_INDEX;
    }
    if (!shader->ready)
    {
        Logger_Log (SCE_INVALID_OPERATION);
        Logger_LogMsg ("you can't use a non-built shader");
        return SCE_SHADER_BAD_INDEX;
    }
    #endif

    return GetIndex[shader->type] (shader, type, name);
}

/*
int SCE_Shader_GetParam(SCE_SShader*, const char*);
float SCE_Shader_GetParamf(SCE_SShader*, const char*);
float* SCE_Shader_GetParamfv(SCE_SShader*, const char*);
*/

static void SCE_Shader_ParamASM (int t, const char *n, int v)
{
    SCE_CSetShaderASMParamf (SCE_Shader_GetIndexASM (used, t, n),
                             (float)v, 0.0, 0.0, 1.0);
}
#ifdef _SCE_USE_CG
static void SCE_Shader_ParamCG (int t, const char *n, int v)
{
    SCE_CSetShaderCGParam ((CGparameter)SCE_Shader_GetIndexCG (used, t, n), v);
}
#endif
static void SCE_Shader_ParamGLSL (int t, const char *n, int v)
{
    SCE_CSetProgramParam (SCE_CGetProgramIndex (used->p_glsl, n), v);
}
/* revise le 19/10/2007 */
void SCE_Shader_Param (int type, const char *name, int val)
{
    Param[used->type] (type, name, val);
}


static void SCE_Shader_ParamfASM (int t, const char *n, float v)
{
    SCE_CSetShaderASMParamf (SCE_Shader_GetIndexASM (used, t, n),
                             v, 0.0, 0.0, 1.0);
}
#ifdef _SCE_USE_CG
static void SCE_Shader_ParamfCG (int t, const char *n, float v)
{
    SCE_CSetShaderCGParamf ((CGparameter)SCE_Shader_GetIndexCG (used, t, n), v);
}
#endif
static void SCE_Shader_ParamfGLSL (int t, const char *n, float v)
{
    SCE_CSetProgramParamf (SCE_CGetProgramIndex (used->p_glsl, n), v);
}
/* revise le 19/10/2007 */
void SCE_Shader_Paramf (int type, const char *name, float val)
{
    Paramf[used->type] (type, name, val);
}


static void SCE_Shader_ParamfvASM (int t, const char *n, size_t s, float *v)
{
    SCE_CSetShaderASMParamfv (SCE_Shader_GetIndexASM (used, t, n), s, v);
}
#ifdef _SCE_USE_CG
static void SCE_Shader_ParamfvCG (int t, const char *n, size_t s, float *v)
{
    SCE_CSetShaderCGParamfv ((CGparameter)SCE_Shader_GetIndexCG (used, t, n),
                             s, v);
}
#endif
static void SCE_Shader_ParamfvGLSL (int t, const char *n, size_t s, float *v)
{
    SCE_CSetProgramParamfv (SCE_CGetProgramIndex (used->p_glsl, n), s, v);
}
/* revise le 19/10/2007 */
void SCE_Shader_Paramfv(int type, const char *name, size_t size, float *val)
{
    Paramfv[used->type] (type, name, size, val);
}


static void SCE_Shader_SetParamASM (int i, int v)
{
    SCE_CSetShaderASMParamf (i, v, 0.0, 0.0, 1.0);
}
#ifdef _SCE_USE_CG
static void SCE_Shader_SetParamCG (int i, int v)
{
    SCE_CSetShaderCGParam ((CGparameter)i, v);
}
#endif
static void SCE_Shader_SetParamGLSL (int i, int v)
{
    SCE_CSetProgramParam (i, v);
}
/* revise le 19/10/2007 */
void SCE_Shader_SetParam (int index, int val)
{
    SetParam[used->type] (index, val);
}


static void SCE_Shader_SetParamfASM (int i, float v)
{
    SCE_CSetShaderASMParamf (i, v, 0.0, 0.0, 1.0);
}
#ifdef _SCE_USE_CG
static void SCE_Shader_SetParamfCG (int i, float v)
{
    SCE_CSetShaderCGParamf ((CGparameter)i, v);
}
#endif
static void SCE_Shader_SetParamfGLSL (int i, float v)
{
    SCE_CSetProgramParamf (i, v);
}
/* revise le 19/10/2007 */
void SCE_Shader_SetParamf (int index, float val)
{
    SetParamf[used->type] (index, val);
}


static void SCE_Shader_SetParamfvASM (int i, size_t s, float *v)
{
    SCE_CSetShaderASMParamfv (i, s, v);
}
#ifdef _SCE_USE_CG
static void SCE_Shader_SetParamfvCG (int i, size_t s, float *v)
{
    SCE_CSetShaderCGParamfv ((CGparameter)i, s, v);
}
#endif
static void SCE_Shader_SetParamfvGLSL (int i, size_t s, float *v)
{
    SCE_CSetProgramParamfv (i, s, v);
}
/* revise le 10/07/2007 */
void SCE_Shader_SetParamfv (int index, size_t size, float *val)
{
    SetParamfv[used->type] (index, size, val);
}


/* ajoute le 14/06/2007 */
void SCE_Shader_Active (int enabled)
{
    sce_shd_enabled = enabled;
}
/* ajoute le 14/06/2007 */
void SCE_Shader_Enable (void)
{
    sce_shd_enabled = SCE_TRUE;
}
/* ajoute le 14/06/2007 */
void SCE_Shader_Disable (void)
{
    sce_shd_enabled = SCE_FALSE;
}


static void SCE_Shader_UseASM (SCE_SShader *shader)
{
    if (!sce_shd_enabled)
        return;
    else if (shader)
    {
        if (shader->v)
            SCE_CUseShaderASM (shader->v);
        if (shader->p)
            SCE_CUseShaderASM (shader->p);
    }
    else
        SCE_CUseShaderASM (NULL);
}
#ifdef _SCE_USE_CG
static void SCE_Shader_UseCG (SCE_SShader *shader)
{
    if (!sce_shd_enabled)
        return;
    else if (shader)
    {
        if (shader->v)
            SCE_CUseShaderCG (shader->v);
        if (shader->p)
            SCE_CUseShaderCG (shader->p);
    }
    else
        SCE_CUseShaderCG (NULL);
}
#endif
/* revise le 12/01/2008 */
static void SCE_Shader_UseGLSL (SCE_SShader *shader)
{
    if (!sce_shd_enabled)
        return;
    else if (shader)
        SCE_CUseProgram (shader->p_glsl);
    else
        SCE_CUseProgram (NULL);
}
/* revise le 19/10/2007 */
void SCE_Shader_Use (SCE_SShader *shader)
{
    static void *pixelshader = NULL;
    static int vs_binded = SCE_FALSE;

    if (shader)
    {
        if (used && used->type != shader->type)
        {
            Use[used->type] (NULL); /* pixel shader */
            Use[used->type] (NULL); /* vertex shader */
            vs_binded = SCE_FALSE;
            pixelshader = NULL;
        }

        if (shader->type == SCE_GLSL_SHADER)
            Use[shader->type] (shader);
        else
        {
            /* analyse du type du shader */
            if (!shader->v)
            {
                /* aucun vertex shader n'est present, on verifie s'il y en
                   a deja un d'actif, si oui, on active le pixel shader
                   du shader envoye, sinon, on retient l'adresse du pixel shader
                   pour une activation ulterieur */
                if (vs_binded)
                    Use[shader->type] (shader);
                else
                    pixelshader = shader->p;
            }
            else
            {
                void *tmp = shader->p;

                if (pixelshader && !tmp)
                {
                    shader->p = pixelshader;
                    pixelshader = NULL;
                }

                /* un vertex shader est present, on l'active */
                Use[shader->type] (shader);

                /* on remet son pixel shader au shader */
                shader->p = tmp;

                /* un vertex shader est binde */
                vs_binded = SCE_TRUE;
            }
        }

        used = shader;
    }
    else
    {
        if (used)
        {
            Use[used->type] (NULL);
            Use[used->type] (NULL);
            used = NULL;
        }

        pixelshader = NULL;
        vs_binded = SCE_FALSE;
    }
}
